from __future__ import annotations

import sys
from datetime import datetime, time
from decimal import Decimal
from typing import Any
from unittest import mock

import pytest
from yarl import URL

from neuro_config_client.entities import (
    AMDGPU,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    BucketsConfig,
    DisksConfig,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GrafanaConfig,
    IdleJobConfig,
    IntelGPU,
    IntelGPUPreset,
    MonitoringConfig,
    NvidiaGPU,
    NvidiaGPUPreset,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchOrchestratorConfigRequest,
    PrometheusConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)
from neuro_config_client.factories import EntityFactory, PayloadFactory

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    # why not backports.zoneinfo: https://github.com/pganssle/zoneinfo/issues/125
    from backports.zoneinfo._zoneinfo import ZoneInfo


@pytest.fixture()
def nvidia_small_gpu() -> str:
    return "nvidia-tesla-k80"


@pytest.fixture()
def amd_small_gpu() -> str:
    return "instinct-mi25"


@pytest.fixture()
def intel_small_gpu() -> str:
    return "flex-170"


class TestEntityFactory:
    @pytest.fixture
    def factory(self) -> EntityFactory:
        return EntityFactory()

    def test_create_cluster_defaults(self, factory: EntityFactory) -> None:
        result = factory.create_cluster(
            {
                "name": "default",
                "status": "blank",
                "orchestrator": {
                    "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                    "job_fallback_hostname": "default.jobs-dev.neu.ro",
                    "job_schedule_timeout_s": 1,
                    "job_schedule_scale_up_timeout_s": 2,
                    "is_http_ingress_secure": False,
                    "resource_pool_types": [{"name": "node-pool"}],
                },
                "storage": {"url": "https://storage-dev.neu.ro"},
                "registry": {
                    "url": "https://registry-dev.neu.ro",
                    "email": "dev@neu.ro",
                },
                "monitoring": {"url": "https://monitoring-dev.neu.ro"},
                "secrets": {"url": "https://secrets-dev.neu.ro"},
                "grafana": {"url": "https://grafana-dev.neu.ro"},
                "prometheus": {"url": "https://prometheus-dev.neu.ro"},
                "disks": {
                    "url": "https://disks-dev.neu.ro",
                    "storage_limit_per_user": 1024,
                },
                "buckets": {
                    "url": "https://buckets-dev.neu.ro",
                    "disable_creation": True,
                },
                "energy": {
                    "co2_grams_eq_per_kwh": 100,
                    "schedules": [
                        {"name": "default", "price_per_kwh": "123.4"},
                        {
                            "name": "green",
                            "price_per_kwh": "123.4",
                            "periods": [
                                {
                                    "weekday": 1,
                                    "start_time": "23:00",
                                    "end_time": "23:59",
                                }
                            ],
                        },
                    ],
                },
                "apps": {
                    "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
                    "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
                },
                "created_at": str(datetime.now()),
            }
        )

        assert result.name == "default"
        assert result.timezone == ZoneInfo("UTC")
        assert result.location is None
        assert result.logo_url is None
        assert result.orchestrator
        assert result.storage
        assert result.registry
        assert result.monitoring
        assert result.secrets
        assert result.grafana
        assert result.prometheus
        assert result.disks
        assert result.dns
        assert result.buckets
        assert result.energy
        assert result.apps
        assert result.created_at

    def test_create_cluster(self, factory: EntityFactory) -> None:
        result = factory.create_cluster(
            {
                "name": "default",
                "status": "blank",
                "location": "us",
                "logo_url": "https://logo",
                "timezone": "America/Los_Angeles",
                "orchestrator": {
                    "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                    "job_fallback_hostname": "default.jobs-dev.neu.ro",
                    "job_schedule_timeout_s": 1,
                    "job_schedule_scale_up_timeout_s": 2,
                    "is_http_ingress_secure": False,
                    "resource_pool_types": [{"name": "node-pool"}],
                },
                "storage": {"url": "https://storage-dev.neu.ro"},
                "registry": {
                    "url": "https://registry-dev.neu.ro",
                    "email": "dev@neu.ro",
                },
                "monitoring": {"url": "https://monitoring-dev.neu.ro"},
                "secrets": {"url": "https://secrets-dev.neu.ro"},
                "grafana": {"url": "https://grafana-dev.neu.ro"},
                "prometheus": {"url": "https://prometheus-dev.neu.ro"},
                "disks": {
                    "url": "https://disks-dev.neu.ro",
                    "storage_limit_per_user": 1024,
                },
                "buckets": {
                    "url": "https://buckets-dev.neu.ro",
                    "disable_creation": True,
                },
                "energy": {
                    "co2_grams_eq_per_kwh": 100,
                    "schedules": [
                        {"name": "default", "price_per_kwh": "123.4"},
                        {
                            "name": "green",
                            "price_per_kwh": "123.4",
                            "periods": [
                                {
                                    "weekday": 1,
                                    "start_time": "23:00",
                                    "end_time": "23:59",
                                }
                            ],
                        },
                    ],
                },
                "apps": {
                    "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
                    "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
                },
                "created_at": str(datetime.now()),
            }
        )

        assert result.name == "default"
        assert result.timezone == ZoneInfo("America/Los_Angeles")
        assert result.location == "us"
        assert result.logo_url == URL("https://logo")

    def test_create_cluster__invalid_timezone(self, factory: EntityFactory) -> None:
        with pytest.raises(ValueError, match="invalid timezone"):
            factory.create_cluster(
                {
                    "name": "default",
                    "status": "blank",
                    "created_at": str(datetime.now()),
                    "timezone": "invalid",
                }
            )

    def test_create_orchestrator(self, factory: EntityFactory) -> None:
        result = factory.create_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
                "resource_pool_types": [{"name": "node-pool"}],
                "resource_presets": [
                    {
                        "name": "cpu-micro",
                        "credits_per_hour": "10",
                        "cpu": 0.1,
                        "memory": 100,
                    }
                ],
                "allow_privileged_mode": True,
                "allow_job_priority": True,
                "pre_pull_images": ["neuromation/base"],
                "idle_jobs": [
                    {
                        "name": "idle",
                        "count": 1,
                        "image": "miner",
                        "resources": {
                            "cpu": 1,
                            "memory": 1024,
                            "nvidia_gpu": 1,
                            "amd_gpu": 2,
                            "intel_gpu": 3,
                        },
                    },
                    {
                        "name": "idle",
                        "count": 1,
                        "image": "miner",
                        "command": ["bash"],
                        "args": ["-c", "sleep infinity"],
                        "resources": {"cpu": 1, "memory": 1024},
                        "env": {"NAME": "VALUE"},
                        "node_selector": {"label": "value"},
                        "image_pull_secret": "secret",
                    },
                ],
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            resource_pool_types=[mock.ANY],
            resource_presets=[mock.ANY],
            allow_privileged_mode=True,
            allow_job_priority=True,
            pre_pull_images=["neuromation/base"],
            idle_jobs=[
                IdleJobConfig(
                    name="idle",
                    count=1,
                    image="miner",
                    resources=Resources(
                        cpu=1,
                        memory=1024,
                        nvidia_gpu=1,
                        amd_gpu=2,
                        intel_gpu=3,
                    ),
                ),
                IdleJobConfig(
                    name="idle",
                    count=1,
                    image="miner",
                    command=["bash"],
                    args=["-c", "sleep infinity"],
                    resources=Resources(cpu=1, memory=1024),
                    env={"NAME": "VALUE"},
                    node_selector={"label": "value"},
                    image_pull_secret="secret",
                ),
            ],
        )

    def test_create_orchestrator_default(self, factory: EntityFactory) -> None:
        result = factory.create_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            resource_pool_types=[],
            resource_presets=[],
            idle_jobs=[],
        )

    def test_create_resource_pool_type(
        self,
        factory: EntityFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
        intel_small_gpu: str,
    ) -> None:
        result = factory.create_resource_pool_type(
            {
                "name": "n1-highmem-4",
                "min_size": 1,
                "max_size": 2,
                "idle_size": 1,
                "cpu": 4.0,
                "available_cpu": 3.0,
                "memory": 12 * 1024,
                "available_memory": 10 * 1024,
                "disk_size": 700 * 2**30,
                "available_disk_size": 600 * 2**30,
                "nvidia_gpu": {
                    "count": 1,
                    "model": nvidia_small_gpu,
                    "memory": 123,
                },
                "nvidia_migs": {
                    "1g.5gb": {
                        "count": 7,
                        "model": f"{nvidia_small_gpu}-1g-5gb",
                        "memory": 456,
                    }
                },
                "amd_gpu": {
                    "count": 2,
                    "model": amd_small_gpu,
                },
                "intel_gpu": {
                    "count": 3,
                    "model": intel_small_gpu,
                },
                "tpu": {
                    "ipv4_cidr_block": "10.0.0.0/8",
                    "types": ["tpu"],
                    "software_versions": ["v1"],
                },
                "is_preemptible": True,
                "price": "1.0",
                "currency": "USD",
                "cpu_min_watts": 1.0,
                "cpu_max_watts": 2.0,
            }
        )

        assert result == ResourcePoolType(
            name="n1-highmem-4",
            min_size=1,
            max_size=2,
            idle_size=1,
            cpu=4.0,
            available_cpu=3.0,
            memory=12 * 1024,
            available_memory=10 * 1024,
            disk_size=700 * 2**30,
            available_disk_size=600 * 2**30,
            nvidia_gpu=NvidiaGPU(
                count=1,
                model=nvidia_small_gpu,
                memory=123,
            ),
            nvidia_migs={
                "1g.5gb": NvidiaGPU(
                    count=7,
                    model=f"{nvidia_small_gpu}-1g-5gb",
                    memory=456,
                )
            },
            amd_gpu=AMDGPU(count=2, model=amd_small_gpu),
            intel_gpu=IntelGPU(count=3, model=intel_small_gpu),
            tpu=mock.ANY,
            is_preemptible=True,
            price=Decimal("1.0"),
            currency="USD",
            cpu_min_watts=1.0,
            cpu_max_watts=2.0,
        )

    def test_create_empty_resource_pool_type(self, factory: EntityFactory) -> None:
        result = factory.create_resource_pool_type({"name": "node-pool"})

        assert result == ResourcePoolType(name="node-pool")

    def test_create_tpu_resource(self, factory: EntityFactory) -> None:
        result = factory.create_tpu_resource(
            {
                "ipv4_cidr_block": "10.0.0.0/8",
                "types": ["tpu"],
                "software_versions": ["v1"],
            }
        )

        assert result == TPUResource(
            ipv4_cidr_block="10.0.0.0/8", types=["tpu"], software_versions=["v1"]
        )

    def test_create_resource_preset_default(self, factory: EntityFactory) -> None:
        result = factory.create_resource_preset(
            {
                "name": "cpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory": 1024,
            }
        )

        assert result == ResourcePreset(
            name="cpu-small", credits_per_hour=Decimal("10"), cpu=4.0, memory=1024
        )

    def test_create_resource_preset_custom(
        self,
        factory: EntityFactory,
        nvidia_small_gpu: str,
    ) -> None:
        result = factory.create_resource_preset(
            {
                "name": "gpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory": 12288,
                "nvidia_gpu": {
                    "count": 1,
                    "model": nvidia_small_gpu,
                    "memory": 123,
                },
                "nvidia_migs": {
                    "1g.5gb": {
                        "count": 7,
                        "model": f"{nvidia_small_gpu}-1g-5gb",
                        "memory": 456,
                    }
                },
                "amd_gpu": {"count": 2},
                "intel_gpu": {"count": 3},
                "tpu": {"type": "tpu", "software_version": "v1"},
                "scheduler_enabled": True,
                "preemptible_node": True,
                "resource_pool_names": ["gpu"],
                "available_resource_pool_names": ["available-gpu"],
                "is_external_job": True,
            }
        )

        assert result == ResourcePreset(
            name="gpu-small",
            credits_per_hour=Decimal("10"),
            cpu=4.0,
            memory=12288,
            nvidia_gpu=NvidiaGPUPreset(
                count=1,
                model=nvidia_small_gpu,
                memory=123,
            ),
            nvidia_migs={
                "1g.5gb": NvidiaGPUPreset(
                    count=7,
                    model=f"{nvidia_small_gpu}-1g-5gb",
                    memory=456,
                )
            },
            amd_gpu=AMDGPUPreset(count=2),
            intel_gpu=IntelGPUPreset(count=3),
            tpu=TPUPreset(type="tpu", software_version="v1"),
            scheduler_enabled=True,
            preemptible_node=True,
            resource_pool_names=["gpu"],
            available_resource_pool_names=["available-gpu"],
            is_external_job=True,
        )

    def test_create_storage(self, factory: EntityFactory) -> None:
        result = factory.create_storage({"url": "https://storage-dev.neu.ro"})

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"), volumes=[]
        )

    def test_create_storage_with_volumes(self, factory: EntityFactory) -> None:
        result = factory.create_storage(
            {
                "url": "https://storage-dev.neu.ro",
                "volumes": [
                    {"name": "test-1"},
                    {
                        "name": "test-2",
                        "path": "/volume",
                        "size": 1024,
                        "credits_per_hour_per_gb": "123",
                    },
                ],
            }
        )

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"),
            volumes=[
                VolumeConfig(name="test-1"),
                VolumeConfig(
                    name="test-2",
                    path="/volume",
                    size=1024,
                    credits_per_hour_per_gb=Decimal(123),
                ),
            ],
        )

    def test_create_registry(self, factory: EntityFactory) -> None:
        result = factory.create_registry({"url": "https://registry-dev.neu.ro"})

        assert result == RegistryConfig(url=URL("https://registry-dev.neu.ro"))

    def test_create_monitoring(self, factory: EntityFactory) -> None:
        result = factory.create_monitoring({"url": "https://monitoring-dev.neu.ro"})

        assert result == MonitoringConfig(url=URL("https://monitoring-dev.neu.ro"))

    def test_create_secrets(self, factory: EntityFactory) -> None:
        result = factory.create_secrets({"url": "https://secrets-dev.neu.ro"})

        assert result == SecretsConfig(url=URL("https://secrets-dev.neu.ro"))

    def test_create_grafana(self, factory: EntityFactory) -> None:
        result = factory.create_grafana({"url": "https://grafana-dev.neu.ro"})

        assert result == GrafanaConfig(
            url=URL("https://grafana-dev.neu.ro"),
        )

    def test_create_prometheus(self, factory: EntityFactory) -> None:
        result = factory.create_prometheus({"url": "https://prometheus-dev.neu.ro"})

        assert result == PrometheusConfig(
            url=URL("https://prometheus-dev.neu.ro"),
        )
        result = factory.create_prometheus({"url": "https://grafana-dev.neu.ro"})

        assert result == PrometheusConfig(
            url=URL("https://prometheus-dev.neu.ro"),
        )

    def test_create_a_record_with_ips(self, factory: EntityFactory) -> None:
        result = factory.create_a_record(
            {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
        )

        assert result == ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])

    def test_create_a_record_dns_name(self, factory: EntityFactory) -> None:
        result = factory.create_a_record(
            {
                "name": "*.jobs-dev.neu.ro.",
                "dns_name": "load-balancer",
                "zone_id": "/hostedzone/1",
                "evaluate_target_health": True,
            }
        )

        assert result == ARecord(
            name="*.jobs-dev.neu.ro.",
            dns_name="load-balancer",
            zone_id="/hostedzone/1",
            evaluate_target_health=True,
        )

    def test_create_disks(self, factory: EntityFactory) -> None:
        result = factory.create_disks(
            {"url": "https://disks-dev.neu.ro", "storage_limit_per_user": 1024}
        )

        assert result == DisksConfig(
            url=URL("https://disks-dev.neu.ro"), storage_limit_per_user=1024
        )

    def test_create_buckets(self, factory: EntityFactory) -> None:
        result = factory.create_buckets(
            {"url": "https://buckets-dev.neu.ro", "disable_creation": True}
        )

        assert result == BucketsConfig(
            url=URL("https://buckets-dev.neu.ro"), disable_creation=True
        )

    def test_create_energy(self, factory: EntityFactory) -> None:
        timezone = ZoneInfo("America/Los_Angeles")
        energy = factory.create_energy(
            {
                "co2_grams_eq_per_kwh": 100,
                "schedules": [
                    {"name": "default", "price_per_kwh": "123.4"},
                    {
                        "name": "green",
                        "price_per_kwh": "123.4",
                        "periods": [
                            {"weekday": 1, "start_time": "23:00", "end_time": "23:59"}
                        ],
                    },
                ],
            },
            timezone=timezone,
        )

        assert energy == EnergyConfig(
            co2_grams_eq_per_kwh=100,
            schedules=[
                EnergySchedule(
                    name="default",
                    price_per_kwh=Decimal("123.4"),
                    periods=[],
                ),
                EnergySchedule(
                    name="green",
                    price_per_kwh=Decimal("123.4"),
                    periods=[
                        EnergySchedulePeriod(
                            weekday=1,
                            start_time=time(hour=23, minute=0, tzinfo=timezone),
                            end_time=time(hour=23, minute=59, tzinfo=timezone),
                        )
                    ],
                ),
            ],
        )

    @pytest.fixture
    def apps(self) -> AppsConfig:
        return AppsConfig(
            apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
            app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
        )

    @pytest.fixture
    def apps_dict(self) -> dict[str, Any]:
        return {
            "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
            "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
        }

    def test_create_apps(
        self, factory: EntityFactory, apps_dict: dict[str, Any], apps: AppsConfig
    ) -> None:
        result = factory.create_apps(apps_dict)
        assert result == apps


class TestPayloadFactory:
    @pytest.fixture
    def factory(self) -> PayloadFactory:
        return PayloadFactory()

    def test_create_patch_cluster_request(self, factory: PayloadFactory) -> None:
        result = factory.create_patch_cluster_request(
            PatchClusterRequest(
                location="us",
                logo_url=URL("https://logo"),
                storage=StorageConfig(url=URL("https://storage-dev.neu.ro")),
                registry=RegistryConfig(url=URL("https://registry-dev.neu.ro")),
                orchestrator=PatchOrchestratorConfigRequest(),
                monitoring=MonitoringConfig(url=URL("https://monitoring-dev.neu.ro")),
                secrets=SecretsConfig(url=URL("https://secrets-dev.neu.ro")),
                grafana=GrafanaConfig(
                    url=URL("https://grafana-dev.neu.ro"),
                ),
                prometheus=PrometheusConfig(
                    url=URL("https://prometheus-dev.neu.ro"),
                ),
                disks=DisksConfig(
                    url=URL("https://disks-dev.neu.ro"), storage_limit_per_user=1024
                ),
                buckets=BucketsConfig(
                    url=URL("https://buckets-dev.neu.ro"), disable_creation=True
                ),
                timezone=ZoneInfo("America/Los_Angeles"),
                energy=EnergyConfig(co2_grams_eq_per_kwh=100),
                apps=AppsConfig(
                    apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
                    app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
                ),
            )
        )

        assert result == {
            "location": "us",
            "logo_url": "https://logo",
            "storage": mock.ANY,
            "registry": mock.ANY,
            "orchestrator": mock.ANY,
            "monitoring": mock.ANY,
            "secrets": mock.ANY,
            "grafana": mock.ANY,
            "prometheus": mock.ANY,
            "disks": mock.ANY,
            "buckets": mock.ANY,
            "timezone": "America/Los_Angeles",
            "energy": mock.ANY,
            "apps": mock.ANY,
        }

    def test_create_patch_cluster_request_default(
        self, factory: PayloadFactory
    ) -> None:
        result = factory.create_patch_cluster_request(PatchClusterRequest())

        assert result == {}

    def test_create_orchestrator(self, factory: PayloadFactory) -> None:
        result = factory.create_orchestrator(
            OrchestratorConfig(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
                allow_privileged_mode=True,
                allow_job_priority=True,
                resource_pool_types=[ResourcePoolType(name="cpu")],
                resource_presets=[
                    ResourcePreset(
                        name="cpu-micro",
                        credits_per_hour=Decimal(10),
                        cpu=0.1,
                        memory=100,
                    )
                ],
                pre_pull_images=["neuromation/base"],
                idle_jobs=[
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        resources=Resources(
                            cpu=1,
                            memory=1024,
                            nvidia_gpu=1,
                            amd_gpu=2,
                            intel_gpu=3,
                        ),
                    ),
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        command=["bash"],
                        args=["-c", "sleep infinity"],
                        resources=Resources(cpu=1, memory=1024),
                        env={"NAME": "VALUE"},
                        node_selector={"label": "value"},
                        image_pull_secret="secret",
                    ),
                ],
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
            "resource_pool_types": [mock.ANY],
            "resource_presets": [mock.ANY],
            "allow_privileged_mode": True,
            "allow_job_priority": True,
            "pre_pull_images": ["neuromation/base"],
            "idle_jobs": [
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "resources": {
                        "cpu": 1,
                        "memory": 1024,
                        "nvidia_gpu": 1,
                        "amd_gpu": 2,
                        "intel_gpu": 3,
                    },
                },
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "command": ["bash"],
                    "args": ["-c", "sleep infinity"],
                    "resources": {"cpu": 1, "memory": 1024},
                    "env": {"NAME": "VALUE"},
                    "node_selector": {"label": "value"},
                    "image_pull_secret": "secret",
                },
            ],
        }

    def test_create_orchestrator_default(self, factory: PayloadFactory) -> None:
        result = factory.create_orchestrator(
            OrchestratorConfig(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
        }

    def test_create_patch_orchestrator_request(self, factory: PayloadFactory) -> None:
        result = factory.create_patch_orchestrator_request(
            PatchOrchestratorConfigRequest(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
                allow_privileged_mode=True,
                allow_job_priority=True,
                resource_pool_types=[ResourcePoolType(name="cpu")],
                resource_presets=[
                    ResourcePreset(
                        name="cpu-micro",
                        credits_per_hour=Decimal(10),
                        cpu=0.1,
                        memory=100,
                    )
                ],
                pre_pull_images=["neuromation/base"],
                idle_jobs=[
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        resources=Resources(cpu=1, memory=1024),
                    )
                ],
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
            "resource_pool_types": [mock.ANY],
            "resource_presets": [mock.ANY],
            "allow_privileged_mode": True,
            "allow_job_priority": True,
            "pre_pull_images": ["neuromation/base"],
            "idle_jobs": [
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "resources": {"cpu": 1, "memory": 1024},
                }
            ],
        }

    def test_create_patch_orchestrator_request_default(
        self, factory: PayloadFactory
    ) -> None:
        result = factory.create_patch_orchestrator_request(
            PatchOrchestratorConfigRequest()
        )

        assert result == {}

    def test_create_resource_pool_type(
        self,
        factory: PayloadFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
        intel_small_gpu: str,
    ) -> None:
        result = factory.create_resource_pool_type(
            ResourcePoolType(
                name="n1-highmem-4",
                min_size=1,
                max_size=2,
                idle_size=1,
                cpu=4.0,
                available_cpu=3.0,
                memory=12 * 1024,
                available_memory=10 * 1024,
                disk_size=700,
                available_disk_size=670,
                nvidia_gpu=NvidiaGPU(
                    count=1,
                    model=nvidia_small_gpu,
                    memory=123,
                ),
                nvidia_migs={
                    "1g.5gb": NvidiaGPU(
                        count=7,
                        model=f"{nvidia_small_gpu}-1g-5gb",
                        memory=456,
                    )
                },
                amd_gpu=AMDGPU(count=2, model=amd_small_gpu),
                intel_gpu=IntelGPU(count=3, model=intel_small_gpu),
                tpu=TPUResource(
                    ipv4_cidr_block="10.0.0.0/8",
                    types=["tpu"],
                    software_versions=["v1"],
                ),
                is_preemptible=True,
                price=Decimal("1.0"),
                currency="USD",
                cpu_min_watts=1.0,
                cpu_max_watts=2.0,
            )
        )

        assert result == {
            "name": "n1-highmem-4",
            "min_size": 1,
            "max_size": 2,
            "idle_size": 1,
            "cpu": 4.0,
            "available_cpu": 3.0,
            "memory": 12 * 1024,
            "available_memory": 10 * 1024,
            "disk_size": 700,
            "available_disk_size": 670,
            "nvidia_gpu": {
                "count": 1,
                "model": nvidia_small_gpu,
                "memory": 123,
            },
            "nvidia_migs": {
                "1g.5gb": {
                    "count": 7,
                    "model": f"{nvidia_small_gpu}-1g-5gb",
                    "memory": 456,
                }
            },
            "amd_gpu": {
                "count": 2,
                "model": amd_small_gpu,
            },
            "intel_gpu": {
                "count": 3,
                "model": intel_small_gpu,
            },
            "tpu": {
                "ipv4_cidr_block": "10.0.0.0/8",
                "types": ["tpu"],
                "software_versions": ["v1"],
            },
            "is_preemptible": True,
            "price": "1.0",
            "currency": "USD",
            "cpu_min_watts": 1.0,
            "cpu_max_watts": 2.0,
        }

    def test_create_empty_resource_pool_type(self, factory: PayloadFactory) -> None:
        result = factory.create_resource_pool_type(ResourcePoolType(name="node-pool"))

        assert result == {
            "name": "node-pool",
            "cpu": 1.0,
            "available_cpu": 1.0,
            "memory": 2**30,
            "available_memory": 2**30,
            "disk_size": 150 * 2**30,
            "available_disk_size": 150 * 2**30,
            "idle_size": 0,
            "is_preemptible": False,
            "min_size": 0,
            "max_size": 1,
        }

    def test_create_tpu_resource(self, factory: PayloadFactory) -> None:
        result = factory.create_tpu_resource(
            TPUResource(
                ipv4_cidr_block="10.0.0.0/8", types=["tpu"], software_versions=["v1"]
            )
        )

        assert result == {
            "ipv4_cidr_block": "10.0.0.0/8",
            "types": ["tpu"],
            "software_versions": ["v1"],
        }

    def test_create_resource_preset(self, factory: PayloadFactory) -> None:
        result = factory.create_resource_preset(
            ResourcePreset(
                name="cpu-small",
                credits_per_hour=Decimal("10"),
                cpu=4.0,
                memory=1024,
            )
        )

        assert result == {
            "name": "cpu-small",
            "credits_per_hour": "10",
            "cpu": 4.0,
            "memory": 1024,
        }

    def test_create_resource_preset_custom(
        self,
        factory: PayloadFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
    ) -> None:
        result = factory.create_resource_preset(
            ResourcePreset(
                name="gpu-small",
                credits_per_hour=Decimal("10"),
                cpu=4.0,
                memory=12288,
                nvidia_gpu=NvidiaGPUPreset(count=1, model=nvidia_small_gpu, memory=123),
                nvidia_migs={
                    "1g.5gb": NvidiaGPUPreset(
                        count=7,
                        model=f"{nvidia_small_gpu}-1g-5gb",
                        memory=456,
                    )
                },
                amd_gpu=AMDGPUPreset(count=2, model=amd_small_gpu),
                intel_gpu=IntelGPUPreset(count=3),
                tpu=TPUPreset(type="tpu", software_version="v1"),
                scheduler_enabled=True,
                preemptible_node=True,
                resource_pool_names=["gpu"],
            )
        )

        assert result == {
            "name": "gpu-small",
            "credits_per_hour": "10",
            "cpu": 4.0,
            "memory": 12288,
            "nvidia_gpu": {
                "count": 1,
                "model": nvidia_small_gpu,
            },
            "nvidia_migs": {
                "1g.5gb": {
                    "count": 7,
                    "model": f"{nvidia_small_gpu}-1g-5gb",
                }
            },
            "amd_gpu": {"count": 2, "model": amd_small_gpu},
            "intel_gpu": {"count": 3},
            "tpu": {"type": "tpu", "software_version": "v1"},
            "scheduler_enabled": True,
            "preemptible_node": True,
            "resource_pool_names": ["gpu"],
        }

    def test_create_storage(self, factory: PayloadFactory) -> None:
        result = factory.create_storage(
            StorageConfig(url=URL("https://storage-dev.neu.ro"))
        )

        assert result == {"url": "https://storage-dev.neu.ro"}

    def test_create_storage_with_volumes(self, factory: PayloadFactory) -> None:
        result = factory.create_storage(
            StorageConfig(
                url=URL("https://storage-dev.neu.ro"),
                volumes=[
                    VolumeConfig(name="test-1"),
                    VolumeConfig(
                        name="test-2",
                        path="/volume",
                        size=1024,
                        credits_per_hour_per_gb=Decimal(123),
                    ),
                ],
            )
        )

        assert result == {
            "url": "https://storage-dev.neu.ro",
            "volumes": [
                {
                    "name": "test-1",
                    "credits_per_hour_per_gb": "0",
                },
                {
                    "name": "test-2",
                    "path": "/volume",
                    "size": 1024,
                    "credits_per_hour_per_gb": "123",
                },
            ],
        }

    def test_create_registry(self, factory: PayloadFactory) -> None:
        result = factory.create_registry(
            RegistryConfig(url=URL("https://registry-dev.neu.ro"))
        )

        assert result == {"url": "https://registry-dev.neu.ro"}

    def test_create_monitoring(self, factory: PayloadFactory) -> None:
        result = factory.create_monitoring(
            MonitoringConfig(url=URL("https://monitoring-dev.neu.ro"))
        )

        assert result == {"url": "https://monitoring-dev.neu.ro"}

    def test_create_secrets(self, factory: PayloadFactory) -> None:
        result = factory.create_secrets(
            SecretsConfig(url=URL("https://secrets-dev.neu.ro"))
        )

        assert result == {"url": "https://secrets-dev.neu.ro"}

    def test_create_grafana(self, factory: PayloadFactory) -> None:
        result = factory.create_grafana(
            GrafanaConfig(
                url=URL("https://grafana-dev.neu.ro"),
            )
        )

        assert result == {
            "url": "https://grafana-dev.neu.ro",
        }

    def test_create_prometheus(self, factory: PayloadFactory) -> None:
        result = factory.create_prometheus(
            PrometheusConfig(
                url=URL("https://prometheus-dev.neu.ro"),
            )
        )

        assert result == {
            "url": "https://prometheus-dev.neu.ro",
        }

    def test_create_a_record_with_ips(self, factory: PayloadFactory) -> None:
        result = factory.create_a_record(
            ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])
        )

        assert result == {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}

    def test_create_a_record_dns_name(self, factory: PayloadFactory) -> None:
        result = factory.create_a_record(
            ARecord(
                name="*.jobs-dev.neu.ro.",
                dns_name="load-balancer",
                zone_id="/hostedzone/1",
                evaluate_target_health=True,
            )
        )

        assert result == {
            "name": "*.jobs-dev.neu.ro.",
            "dns_name": "load-balancer",
            "zone_id": "/hostedzone/1",
            "evaluate_target_health": True,
        }

    def test_create_disks(self, factory: PayloadFactory) -> None:
        result = factory.create_disks(
            DisksConfig(
                url=URL("https://disks-dev.neu.ro"), storage_limit_per_user=1024
            )
        )

        assert result == {
            "url": "https://disks-dev.neu.ro",
            "storage_limit_per_user": 1024,
        }

    def test_create_buckets(self, factory: PayloadFactory) -> None:
        result = factory.create_buckets(
            BucketsConfig(url=URL("https://buckets-dev.neu.ro"), disable_creation=True)
        )

        assert result == {"url": "https://buckets-dev.neu.ro", "disable_creation": True}

    def test_create_energy(self, factory: PayloadFactory) -> None:
        timezone = ZoneInfo("America/Los_Angeles")
        energy = factory.create_energy(
            EnergyConfig(
                co2_grams_eq_per_kwh=100,
                schedules=[
                    EnergySchedule(name="default", price_per_kwh=Decimal("246.8")),
                    EnergySchedule(
                        name="green",
                        price_per_kwh=Decimal("123.4"),
                        periods=[
                            EnergySchedulePeriod(
                                weekday=1,
                                start_time=time(hour=23, minute=0, tzinfo=timezone),
                                end_time=time(hour=0, minute=0, tzinfo=timezone),
                            )
                        ],
                    ),
                ],
            )
        )

        assert energy == {
            "co2_grams_eq_per_kwh": 100,
            "schedules": [
                {"name": "default", "price_per_kwh": "246.8"},
                {
                    "name": "green",
                    "price_per_kwh": "123.4",
                    "periods": [
                        {"weekday": 1, "start_time": "23:00", "end_time": "00:00"}
                    ],
                },
            ],
        }

    @pytest.fixture
    def apps(self) -> AppsConfig:
        return AppsConfig(
            apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
            app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
        )

    def test_create_apps(self, factory: PayloadFactory, apps: AppsConfig) -> None:
        result = factory.create_apps(apps)
        assert result == {
            "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
            "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
        }



from __future__ import annotations

import sys
from datetime import datetime, time, tzinfo
from decimal import Decimal
from typing import Any

from yarl import URL

from .entities import (
    AMDGPU,
    GPU,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    BucketsConfig,
    Cluster,
    DisksConfig,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GPUPreset,
    GrafanaConfig,
    IdleJobConfig,
    IntelGPU,
    IntelGPUPreset,
    MonitoringConfig,
    NvidiaGPU,
    NvidiaGPUPreset,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchOrchestratorConfigRequest,
    PrometheusConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    # why not backports.zoneinfo: https://github.com/pganssle/zoneinfo/issues/125
    from backports.zoneinfo._zoneinfo import ZoneInfo


class EntityFactory:
    def create_cluster(self, payload: dict[str, Any]) -> Cluster:
        timezone = self._create_timezone(payload.get("timezone"))
        return Cluster(
            name=payload["name"],
            location=payload.get("location"),
            logo_url=URL(logo_url) if (logo_url := payload.get("logo_url")) else None,
            orchestrator=self.create_orchestrator(payload["orchestrator"]),
            storage=self.create_storage(payload["storage"]),
            registry=self.create_registry(payload["registry"]),
            monitoring=self.create_monitoring(payload["monitoring"]),
            secrets=self.create_secrets(payload["secrets"]),
            grafana=self.create_grafana(payload.get("grafana") or payload["metrics"]),
            prometheus=self.create_prometheus(
                payload.get("prometheus") or payload["metrics"]
            ),
            disks=self.create_disks(payload["disks"]),
            buckets=self.create_buckets(payload["buckets"]),
            created_at=datetime.fromisoformat(payload["created_at"]),
            timezone=timezone,
            energy=self.create_energy(payload["energy"], timezone=timezone),
            apps=self.create_apps(payload["apps"]),
        )

    def create_orchestrator(self, payload: dict[str, Any]) -> OrchestratorConfig:
        return OrchestratorConfig(
            job_hostname_template=payload["job_hostname_template"],
            job_fallback_hostname=payload["job_fallback_hostname"],
            job_schedule_timeout_s=payload["job_schedule_timeout_s"],
            job_schedule_scale_up_timeout_s=payload["job_schedule_scale_up_timeout_s"],
            is_http_ingress_secure=payload["is_http_ingress_secure"],
            resource_pool_types=[
                self.create_resource_pool_type(r)
                for r in payload.get("resource_pool_types", ())
            ],
            resource_presets=[
                self.create_resource_preset(preset)
                for preset in payload.get("resource_presets", ())
            ],
            allow_privileged_mode=payload.get(
                "allow_privileged_mode", OrchestratorConfig.allow_privileged_mode
            ),
            allow_job_priority=payload.get(
                "allow_job_priority", OrchestratorConfig.allow_job_priority
            ),
            pre_pull_images=payload.get("pre_pull_images", ()),
            idle_jobs=[
                self.create_idle_job(job) for job in payload.get("idle_jobs", ())
            ],
        )

    def create_resource_pool_type(self, payload: dict[str, Any]) -> ResourcePoolType:
        cpu = payload.get("cpu", ResourcePoolType.cpu)
        memory = payload.get("memory", ResourcePoolType.memory)
        disk_size = payload.get("disk_size", ResourcePoolType.disk_size)
        return ResourcePoolType(
            name=payload["name"],
            min_size=payload.get("min_size", ResourcePoolType.min_size),
            max_size=payload.get("max_size", ResourcePoolType.max_size),
            idle_size=payload.get("idle_size", ResourcePoolType.idle_size),
            cpu=cpu,
            available_cpu=payload.get("available_cpu") or cpu,
            memory=memory,
            available_memory=payload.get("available_memory") or memory,
            disk_size=disk_size,
            available_disk_size=payload.get("available_disk_size", disk_size),
            nvidia_gpu=(
                self._create_nvidia_gpu(nvidia_gpu)
                if (nvidia_gpu := payload.get("nvidia_gpu"))
                else None
            ),
            nvidia_migs=(
                {key: self._create_nvidia_gpu(value) for key, value in migs.items()}
                if (migs := payload.get("nvidia_migs"))
                else None
            ),
            amd_gpu=(
                self._create_amd_gpu(amd_gpu)
                if (amd_gpu := payload.get("amd_gpu"))
                else None
            ),
            intel_gpu=(
                self._create_intel_gpu(intel_gpu)
                if (intel_gpu := payload.get("intel_gpu"))
                else None
            ),
            tpu=(
                self.create_tpu_resource(tpu) if (tpu := payload.get("tpu")) else None
            ),
            price=Decimal(payload.get("price", ResourcePoolType.price)),
            currency=payload.get("currency"),
            is_preemptible=payload.get(
                "is_preemptible", ResourcePoolType.is_preemptible
            ),
            cpu_min_watts=payload.get("cpu_min_watts", ResourcePoolType.cpu_min_watts),
            cpu_max_watts=payload.get("cpu_max_watts", ResourcePoolType.cpu_max_watts),
        )

    def _create_nvidia_gpu(self, payload: dict[str, Any]) -> NvidiaGPU:
        return NvidiaGPU(
            count=payload["count"],
            model=payload["model"],
            memory=payload.get("memory"),
        )

    def _create_amd_gpu(self, payload: dict[str, Any]) -> AMDGPU:
        return AMDGPU(
            count=payload["count"],
            model=payload["model"],
            memory=payload.get("memory"),
        )

    def _create_intel_gpu(self, payload: dict[str, Any]) -> IntelGPU:
        return IntelGPU(
            count=payload["count"],
            model=payload["model"],
            memory=payload.get("memory"),
        )

    def create_tpu_resource(self, payload: dict[str, Any]) -> TPUResource:
        return TPUResource(
            ipv4_cidr_block=payload["ipv4_cidr_block"],
            types=list(payload["types"]),
            software_versions=list(payload["software_versions"]),
        )

    def create_resource_preset(self, payload: dict[str, Any]) -> ResourcePreset:
        return ResourcePreset(
            name=payload["name"],
            credits_per_hour=Decimal(payload["credits_per_hour"]),
            cpu=payload["cpu"],
            memory=payload["memory"],
            nvidia_gpu=(
                self._create_nvidia_gpu_preset(nvidia_gpu)
                if (nvidia_gpu := payload.get("nvidia_gpu"))
                else None
            ),
            nvidia_migs=(
                {
                    key: self._create_nvidia_gpu_preset(value)
                    for key, value in migs.items()
                }
                if (migs := payload.get("nvidia_migs"))
                else None
            ),
            amd_gpu=(
                self._create_amd_gpu_preset(amd_gpu)
                if (amd_gpu := payload.get("amd_gpu"))
                else None
            ),
            intel_gpu=(
                self._create_intel_gpu_preset(intel_gpu)
                if (intel_gpu := payload.get("intel_gpu"))
                else None
            ),
            tpu=self.create_tpu_preset(tpu) if (tpu := payload.get("tpu")) else None,
            scheduler_enabled=payload.get("scheduler_enabled", False),
            preemptible_node=payload.get("preemptible_node", False),
            is_external_job=payload.get("is_external_job", False),
            resource_pool_names=payload.get("resource_pool_names", ()),
            available_resource_pool_names=payload.get(
                "available_resource_pool_names", ()
            ),
        )

    def _create_nvidia_gpu_preset(self, payload: dict[str, Any]) -> NvidiaGPUPreset:
        return NvidiaGPUPreset(
            count=payload["count"],
            model=payload.get("model"),
            memory=payload.get("memory"),
        )

    def _create_amd_gpu_preset(self, payload: dict[str, Any]) -> AMDGPUPreset:
        return AMDGPUPreset(
            count=payload["count"],
            model=payload.get("model"),
            memory=payload.get("memory"),
        )

    def _create_intel_gpu_preset(self, payload: dict[str, Any]) -> IntelGPUPreset:
        return IntelGPUPreset(
            count=payload["count"],
            model=payload.get("model"),
            memory=payload.get("memory"),
        )

    def create_tpu_preset(self, payload: dict[str, Any]) -> TPUPreset:
        return TPUPreset(
            type=payload["type"], software_version=payload["software_version"]
        )

    def create_idle_job(self, payload: dict[str, Any]) -> IdleJobConfig:
        return IdleJobConfig(
            name=payload["name"],
            count=payload["count"],
            image=payload["image"],
            command=payload.get("command", []),
            args=payload.get("args", []),
            image_pull_secret=payload.get("image_pull_secret"),
            resources=self.create_resources(payload["resources"]),
            env=payload.get("env") or {},
            node_selector=payload.get("node_selector") or {},
        )

    def create_resources(self, payload: dict[str, Any]) -> Resources:
        return Resources(
            cpu=payload["cpu"],
            memory=payload["memory"],
            nvidia_gpu=payload.get("nvidia_gpu", 0),
            amd_gpu=payload.get("amd_gpu", 0),
            intel_gpu=payload.get("intel_gpu", 0),
        )

    def create_storage(self, payload: dict[str, Any]) -> StorageConfig:
        return StorageConfig(
            url=URL(payload["url"]),
            volumes=[self.create_volume(e) for e in payload.get("volumes", ())],
        )

    def create_volume(self, payload: dict[str, Any]) -> VolumeConfig:
        return VolumeConfig(
            name=payload["name"],
            path=payload.get("path"),
            size=payload.get("size"),
            credits_per_hour_per_gb=Decimal(
                payload.get(
                    "credits_per_hour_per_gb", VolumeConfig.credits_per_hour_per_gb
                )
            ),
        )

    def create_registry(self, payload: dict[str, Any]) -> RegistryConfig:
        return RegistryConfig(url=URL(payload["url"]))

    def create_monitoring(self, payload: dict[str, Any]) -> MonitoringConfig:
        return MonitoringConfig(url=URL(payload["url"]))

    def create_secrets(self, payload: dict[str, Any]) -> SecretsConfig:
        return SecretsConfig(url=URL(payload["url"]))

    def create_grafana(self, payload: dict[str, Any]) -> GrafanaConfig:
        return GrafanaConfig(url=URL(payload["url"]))

    def create_prometheus(self, payload: dict[str, Any]) -> PrometheusConfig:
        return PrometheusConfig(
            url=URL(payload["url"].replace("grafana", "prometheus"))
        )

    def create_a_record(self, payload: dict[str, Any]) -> ARecord:
        return ARecord(
            name=payload["name"],
            ips=payload.get("ips", ()),
            dns_name=payload.get("dns_name", ARecord.dns_name),
            zone_id=payload.get("zone_id", ARecord.zone_id),
            evaluate_target_health=payload.get(
                "evaluate_target_health", ARecord.evaluate_target_health
            ),
        )

    def create_disks(self, payload: dict[str, Any]) -> DisksConfig:
        return DisksConfig(
            url=URL(payload["url"]),
            storage_limit_per_user=payload["storage_limit_per_user"],
        )

    def create_buckets(self, payload: dict[str, Any]) -> BucketsConfig:
        return BucketsConfig(
            url=URL(payload["url"]),
            disable_creation=payload.get("disable_creation", False),
        )

    def _create_timezone(self, name: str | None) -> tzinfo:
        if not name:
            return Cluster.timezone
        try:
            return ZoneInfo(name)
        except Exception:
            raise ValueError(f"invalid timezone: {name}")

    def create_energy(
        self, payload: dict[str, Any], *, timezone: tzinfo
    ) -> EnergyConfig:
        return EnergyConfig(
            co2_grams_eq_per_kwh=payload["co2_grams_eq_per_kwh"],
            schedules=[
                self._create_energy_schedule(s, timezone)
                for s in payload.get("schedules", ())
            ],
        )

    def _create_energy_schedule(
        self, payload: dict[str, Any], timezone: tzinfo
    ) -> EnergySchedule:
        return EnergySchedule(
            name=payload["name"],
            price_per_kwh=Decimal(payload["price_per_kwh"]),
            periods=[
                self._create_energy_schedule_period(p, timezone=timezone)
                for p in payload.get("periods", ())
            ],
        )

    def _create_energy_schedule_period(
        self, payload: dict[str, Any], *, timezone: tzinfo
    ) -> EnergySchedulePeriod:
        start_time = time.fromisoformat(payload["start_time"]).replace(tzinfo=timezone)
        end_time = time.fromisoformat(payload["end_time"]).replace(tzinfo=timezone)
        return EnergySchedulePeriod(
            weekday=payload["weekday"],
            start_time=start_time,
            end_time=end_time,
        )

    def create_apps(self, payload: dict[str, Any]) -> AppsConfig:
        return AppsConfig(
            apps_hostname_templates=payload.get("apps_hostname_templates", []),
            app_proxy_url=URL(payload["app_proxy_url"]),
        )


class PayloadFactory:
    @classmethod
    def create_patch_cluster_request(
        cls, request: PatchClusterRequest
    ) -> dict[str, Any]:
        payload: dict[str, Any] = {}
        if request.location:
            payload["location"] = request.location
        if request.logo_url:
            payload["logo_url"] = str(request.logo_url)
        if request.storage:
            payload["storage"] = cls.create_storage(request.storage)
        if request.registry:
            payload["registry"] = cls.create_registry(request.registry)
        if request.orchestrator:
            payload["orchestrator"] = cls.create_patch_orchestrator_request(
                request.orchestrator
            )
        if request.monitoring:
            payload["monitoring"] = cls.create_monitoring(request.monitoring)
        if request.secrets:
            payload["secrets"] = cls.create_secrets(request.secrets)
        if request.grafana:
            payload["grafana"] = cls.create_grafana(request.grafana)
        if request.prometheus:
            payload["prometheus"] = cls.create_prometheus(request.prometheus)
        if request.disks:
            payload["disks"] = cls.create_disks(request.disks)
        if request.buckets:
            payload["buckets"] = cls.create_buckets(request.buckets)
        if request.timezone:
            payload["timezone"] = str(request.timezone)
        if request.energy:
            payload["energy"] = cls.create_energy(request.energy)
        if request.apps:
            payload["apps"] = cls.create_apps(request.apps)
        return payload

    @classmethod
    def create_storage(cls, storage: StorageConfig) -> dict[str, Any]:
        result: dict[str, Any] = {"url": str(storage.url)}
        if storage.volumes:
            result["volumes"] = [cls._create_volume(e) for e in storage.volumes]
        return result

    @classmethod
    def _create_volume(cls, volume: VolumeConfig) -> dict[str, Any]:
        result: dict[str, Any] = {
            "name": volume.name,
            "credits_per_hour_per_gb": str(volume.credits_per_hour_per_gb),
        }
        if volume.path:
            result["path"] = volume.path
        if volume.size is not None:
            result["size"] = volume.size
        return result

    @classmethod
    def create_registry(cls, registry: RegistryConfig) -> dict[str, Any]:
        return {"url": str(registry.url)}

    def create_orchestrator(self, orchestrator: OrchestratorConfig) -> dict[str, Any]:
        result = {
            "job_hostname_template": orchestrator.job_hostname_template,
            "is_http_ingress_secure": orchestrator.is_http_ingress_secure,
            "job_fallback_hostname": orchestrator.job_fallback_hostname,
            "job_schedule_timeout_s": orchestrator.job_schedule_timeout_s,
            "job_schedule_scale_up_timeout_s": (
                orchestrator.job_schedule_scale_up_timeout_s
            ),
        }
        if orchestrator.allow_privileged_mode:
            result["allow_privileged_mode"] = orchestrator.allow_privileged_mode
        if orchestrator.allow_job_priority:
            result["allow_job_priority"] = orchestrator.allow_job_priority
        if orchestrator.resource_pool_types:
            result["resource_pool_types"] = [
                self.create_resource_pool_type(r)
                for r in orchestrator.resource_pool_types
            ]
        if orchestrator.resource_presets:
            result["resource_presets"] = [
                self.create_resource_preset(preset)
                for preset in orchestrator.resource_presets
            ]
        if orchestrator.pre_pull_images:
            result["pre_pull_images"] = orchestrator.pre_pull_images
        if orchestrator.idle_jobs:
            result["idle_jobs"] = [
                self._create_idle_job(job) for job in orchestrator.idle_jobs
            ]
        return result

    @classmethod
    def create_patch_orchestrator_request(
        cls, orchestrator: PatchOrchestratorConfigRequest
    ) -> dict[str, Any]:
        payload: dict[str, Any] = {}
        if orchestrator.job_hostname_template:
            payload["job_hostname_template"] = orchestrator.job_hostname_template
        if orchestrator.is_http_ingress_secure is not None:
            payload["is_http_ingress_secure"] = orchestrator.is_http_ingress_secure
        if orchestrator.job_fallback_hostname:
            payload["job_fallback_hostname"] = orchestrator.job_fallback_hostname
        if orchestrator.job_schedule_timeout_s is not None:
            payload["job_schedule_timeout_s"] = orchestrator.job_schedule_timeout_s
        if orchestrator.job_schedule_scale_up_timeout_s is not None:
            payload["job_schedule_scale_up_timeout_s"] = (
                orchestrator.job_schedule_scale_up_timeout_s
            )
        if orchestrator.allow_privileged_mode is not None:
            payload["allow_privileged_mode"] = orchestrator.allow_privileged_mode
        if orchestrator.allow_job_priority is not None:
            payload["allow_job_priority"] = orchestrator.allow_job_priority
        if orchestrator.resource_pool_types:
            payload["resource_pool_types"] = [
                cls.create_resource_pool_type(r)
                for r in orchestrator.resource_pool_types
            ]
        if orchestrator.resource_presets:
            payload["resource_presets"] = [
                cls.create_resource_preset(preset)
                for preset in orchestrator.resource_presets
            ]
        if orchestrator.pre_pull_images:
            payload["pre_pull_images"] = orchestrator.pre_pull_images
        if orchestrator.idle_jobs:
            payload["idle_jobs"] = [
                cls._create_idle_job(job) for job in orchestrator.idle_jobs
            ]
        return payload

    @classmethod
    def create_resource_pool_type(
        cls, resource_pool_type: ResourcePoolType
    ) -> dict[str, Any]:
        result: dict[str, Any] = {
            "name": resource_pool_type.name,
            "is_preemptible": resource_pool_type.is_preemptible,
            "min_size": resource_pool_type.min_size,
            "max_size": resource_pool_type.max_size,
            "idle_size": resource_pool_type.idle_size,
            "cpu": resource_pool_type.cpu,
            "available_cpu": resource_pool_type.available_cpu,
            "memory": resource_pool_type.memory,
            "available_memory": resource_pool_type.available_memory,
            "disk_size": resource_pool_type.disk_size,
            "available_disk_size": resource_pool_type.available_disk_size,
        }
        if resource_pool_type.nvidia_gpu:
            result["nvidia_gpu"] = cls._create_gpu(resource_pool_type.nvidia_gpu)
        if resource_pool_type.nvidia_migs:
            result["nvidia_migs"] = {
                key: cls._create_gpu(value)
                for key, value in resource_pool_type.nvidia_migs.items()
            }
        if resource_pool_type.amd_gpu:
            result["amd_gpu"] = cls._create_gpu(resource_pool_type.amd_gpu)
        if resource_pool_type.intel_gpu:
            result["intel_gpu"] = cls._create_gpu(resource_pool_type.intel_gpu)
        if resource_pool_type.currency:
            result["price"] = str(resource_pool_type.price)
            result["currency"] = resource_pool_type.currency
        if resource_pool_type.tpu:
            result["tpu"] = cls.create_tpu_resource(resource_pool_type.tpu)
        if resource_pool_type.cpu_min_watts:
            result["cpu_min_watts"] = resource_pool_type.cpu_min_watts
        if resource_pool_type.cpu_max_watts:
            result["cpu_max_watts"] = resource_pool_type.cpu_max_watts
        return result

    @classmethod
    def _create_gpu(cls, gpu: GPU) -> dict[str, Any]:
        result: dict[str, Any] = {
            "count": gpu.count,
            "model": gpu.model,
        }
        if gpu.memory:
            result["memory"] = gpu.memory
        return result

    @classmethod
    def create_tpu_resource(cls, tpu: TPUResource) -> dict[str, Any]:
        return {
            "ipv4_cidr_block": tpu.ipv4_cidr_block,
            "types": list(tpu.types),
            "software_versions": list(tpu.software_versions),
        }

    @classmethod
    def create_resource_preset(cls, preset: ResourcePreset) -> dict[str, Any]:
        result: dict[str, Any] = {
            "name": preset.name,
            "credits_per_hour": str(preset.credits_per_hour),
            "cpu": preset.cpu,
            "memory": preset.memory,
        }
        if preset.nvidia_gpu:
            result["nvidia_gpu"] = cls._create_gpu_preset(preset.nvidia_gpu)
        if preset.nvidia_migs:
            result["nvidia_migs"] = {
                key: cls._create_gpu_preset(value)
                for key, value in preset.nvidia_migs.items()
            }
        if preset.amd_gpu:
            result["amd_gpu"] = cls._create_gpu_preset(preset.amd_gpu)
        if preset.intel_gpu:
            result["intel_gpu"] = cls._create_gpu_preset(preset.intel_gpu)
        if preset.tpu:
            result["tpu"] = cls._create_tpu_preset(preset.tpu)
        if preset.scheduler_enabled:
            result["scheduler_enabled"] = preset.scheduler_enabled
        if preset.preemptible_node:
            result["preemptible_node"] = preset.preemptible_node
        if preset.resource_pool_names:
            result["resource_pool_names"] = preset.resource_pool_names
        return result

    @classmethod
    def _create_gpu_preset(cls, gpu_preset: GPUPreset) -> dict[str, Any]:
        result: dict[str, Any] = {"count": gpu_preset.count}
        if gpu_preset.model:
            result["model"] = gpu_preset.model
        return result

    @classmethod
    def _create_tpu_preset(cls, tpu: TPUPreset) -> dict[str, Any]:
        return {"type": tpu.type, "software_version": tpu.software_version}

    @classmethod
    def _create_idle_job(cls, idle_job: IdleJobConfig) -> dict[str, Any]:
        result = {
            "name": idle_job.name,
            "count": idle_job.count,
            "image": idle_job.image,
            "resources": cls._create_resources(idle_job.resources),
        }
        if idle_job.command:
            result["command"] = idle_job.command
        if idle_job.args:
            result["args"] = idle_job.args
        if idle_job.image_pull_secret:
            result["image_pull_secret"] = idle_job.image_pull_secret
        if idle_job.env:
            result["env"] = idle_job.env
        if idle_job.node_selector:
            result["node_selector"] = idle_job.node_selector
        return result

    @classmethod
    def _create_resources(cls, resources: Resources) -> dict[str, Any]:
        result = {"cpu": resources.cpu, "memory": resources.memory}
        if resources.nvidia_gpu:
            result["nvidia_gpu"] = resources.nvidia_gpu
        if resources.amd_gpu:
            result["amd_gpu"] = resources.amd_gpu
        if resources.intel_gpu:
            result["intel_gpu"] = resources.intel_gpu
        return result

    @classmethod
    def create_monitoring(cls, monitoring: MonitoringConfig) -> dict[str, Any]:
        return {"url": str(monitoring.url)}

    @classmethod
    def create_grafana(cls, grafana: GrafanaConfig) -> dict[str, Any]:
        return {"url": str(grafana.url)}

    @classmethod
    def create_prometheus(cls, prometheus: PrometheusConfig) -> dict[str, Any]:
        return {"url": str(prometheus.url)}

    @classmethod
    def create_secrets(cls, secrets: SecretsConfig) -> dict[str, Any]:
        return {"url": str(secrets.url)}

    @classmethod
    def create_buckets(cls, buckets: BucketsConfig) -> dict[str, Any]:
        return {"url": str(buckets.url), "disable_creation": buckets.disable_creation}

    @classmethod
    def create_a_record(cls, a_record: ARecord) -> dict[str, Any]:
        result: dict[str, Any] = {
            "name": a_record.name,
        }
        if a_record.ips:
            result["ips"] = a_record.ips
        if a_record.dns_name:
            result["dns_name"] = a_record.dns_name
        if a_record.zone_id:
            result["zone_id"] = a_record.zone_id
        if a_record.evaluate_target_health:
            result["evaluate_target_health"] = a_record.evaluate_target_health
        return result

    @classmethod
    def create_disks(cls, disks: DisksConfig) -> dict[str, Any]:
        return {
            "url": str(disks.url),
            "storage_limit_per_user": disks.storage_limit_per_user,
        }

    @classmethod
    def create_energy(cls, energy: EnergyConfig) -> dict[str, Any]:
        return {
            "co2_grams_eq_per_kwh": energy.co2_grams_eq_per_kwh,
            "schedules": [cls._create_energy_schedule(s) for s in energy.schedules],
        }

    @classmethod
    def _create_energy_schedule(cls, schedule: EnergySchedule) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "name": schedule.name,
            "price_per_kwh": str(schedule.price_per_kwh),
        }
        periods = [cls._create_energy_schedule_period(p) for p in schedule.periods]
        if periods:
            payload["periods"] = periods
        return payload

    @classmethod
    def _create_energy_schedule_period(
        cls, period: EnergySchedulePeriod
    ) -> dict[str, Any]:
        return {
            "weekday": period.weekday,
            "start_time": period.start_time.strftime("%H:%M"),
            "end_time": period.end_time.strftime("%H:%M"),
        }

    @classmethod
    def create_apps(cls, apps: AppsConfig) -> dict[str, Any]:
        return {
            "apps_hostname_templates": apps.apps_hostname_templates,
            "app_proxy_url": str(apps.app_proxy_url),
        }



"""Platform config client."""

from importlib.metadata import version

from .client import ConfigClient, ConfigClientBase
from .entities import (
    AMDGPU,
    GPU,
    ACMEEnvironment,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    BucketsConfig,
    Cluster,
    DisksConfig,
    DNSConfig,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GPUPreset,
    GrafanaConfig,
    IdleJobConfig,
    IngressConfig,
    IntelGPU,
    IntelGPUPreset,
    MonitoringConfig,
    NvidiaGPU,
    NvidiaGPUPreset,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchOrchestratorConfigRequest,
    PrometheusConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)

__all__ = [
    "ConfigClient",
    "ConfigClientBase",
    "AMDGPU",
    "GPU",
    "ACMEEnvironment",
    "AddNodePoolRequest",
    "AMDGPUPreset",
    "AppsConfig",
    "ARecord",
    "BucketsConfig",
    "Cluster",
    "DisksConfig",
    "DNSConfig",
    "EnergyConfig",
    "EnergySchedule",
    "EnergySchedulePeriod",
    "GPUPreset",
    "GrafanaConfig",
    "IdleJobConfig",
    "IngressConfig",
    "IntelGPU",
    "IntelGPUPreset",
    "MonitoringConfig",
    "NvidiaGPU",
    "NvidiaGPUPreset",
    "OrchestratorConfig",
    "PatchClusterRequest",
    "PatchOrchestratorConfigRequest",
    "PrometheusConfig",
    "RegistryConfig",
    "ResourcePoolType",
    "ResourcePreset",
    "Resources",
    "SecretsConfig",
    "StorageConfig",
    "TPUPreset",
    "TPUResource",
    "VolumeConfig",
]
__version__ = version(__name__)



from __future__ import annotations

import enum
import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from datetime import datetime, time, tzinfo
from decimal import Decimal

from yarl import URL

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    # why not backports.zoneinfo: https://github.com/pganssle/zoneinfo/issues/125
    from backports.zoneinfo._zoneinfo import ZoneInfo

UTC = ZoneInfo("UTC")

UNSET_DATETIME = datetime.min.replace(tzinfo=UTC)


@dataclass(frozen=True)
class VolumeConfig:
    name: str
    size: int | None = None
    path: str | None = None
    credits_per_hour_per_gb: Decimal = Decimal(0)


@dataclass(frozen=True)
class StorageConfig:
    url: URL
    volumes: Sequence[VolumeConfig] = ()


@dataclass(frozen=True)
class RegistryConfig:
    url: URL

    @property
    def host(self) -> str:
        """Returns registry hostname with port (if specified)"""
        port = self.url.explicit_port
        suffix = f":{port}" if port is not None else ""
        return f"{self.url.host}{suffix}"


@dataclass(frozen=True)
class MonitoringConfig:
    url: URL


@dataclass(frozen=True)
class GrafanaConfig:
    url: URL


@dataclass(frozen=True)
class PrometheusConfig:
    url: URL


@dataclass(frozen=True)
class SecretsConfig:
    url: URL


@dataclass(frozen=True)
class DisksConfig:
    url: URL
    storage_limit_per_user: int = 500 * 2**30  # 500gb


@dataclass(frozen=True)
class BucketsConfig:
    url: URL
    disable_creation: bool = False


class ACMEEnvironment(str, enum.Enum):
    STAGING = "staging"
    PRODUCTION = "production"


@dataclass(frozen=True)
class IngressConfig:
    acme_environment: ACMEEnvironment = ACMEEnvironment.PRODUCTION
    default_cors_origins: Sequence[str] = ()
    additional_cors_origins: Sequence[str] = ()


@dataclass(frozen=True)
class GPUPreset:
    count: int
    model: str | None = None
    memory: int | None = None


@dataclass(frozen=True)
class NvidiaGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class AMDGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class IntelGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class TPUPreset:
    type: str
    software_version: str


@dataclass(frozen=True)
class ResourcePreset:
    name: str
    credits_per_hour: Decimal
    cpu: float
    memory: int
    nvidia_gpu: NvidiaGPUPreset | None = None
    nvidia_migs: Mapping[str, NvidiaGPUPreset] | None = None
    amd_gpu: AMDGPUPreset | None = None
    intel_gpu: IntelGPUPreset | None = None
    tpu: TPUPreset | None = None
    scheduler_enabled: bool = False
    preemptible_node: bool = False
    is_external_job: bool = False
    resource_pool_names: Sequence[str] = ()
    available_resource_pool_names: Sequence[str] = ()


@dataclass(frozen=True)
class GPU:
    count: int
    model: str
    memory: int | None = None


@dataclass(frozen=True)
class NvidiaGPU(GPU):
    pass


@dataclass(frozen=True)
class AMDGPU(GPU):
    pass


@dataclass(frozen=True)
class IntelGPU(GPU):
    pass


@dataclass(frozen=True)
class TPUResource:
    ipv4_cidr_block: str
    types: Sequence[str] = ()
    software_versions: Sequence[str] = ()


@dataclass(frozen=True)
class ResourcePoolType:
    name: str
    min_size: int = 0
    max_size: int = 1
    idle_size: int = 0

    cpu: float = 1.0
    available_cpu: float = 1.0
    memory: int = 2**30  # 1gb
    available_memory: int = 2**30
    disk_size: int = 150 * 2**30  # 150gb
    available_disk_size: int = 150 * 2**30  # 150gb

    nvidia_gpu: NvidiaGPU | None = None
    nvidia_migs: Mapping[str, NvidiaGPU] | None = None
    amd_gpu: AMDGPU | None = None
    intel_gpu: IntelGPU | None = None
    tpu: TPUResource | None = None

    price: Decimal = Decimal()
    currency: str | None = None

    is_preemptible: bool = False

    cpu_min_watts: float = 0.0
    cpu_max_watts: float = 0.0

    @property
    def has_gpu(self) -> bool:
        return any((self.nvidia_gpu, self.amd_gpu, self.intel_gpu))


@dataclass(frozen=True)
class Resources:
    cpu: float
    memory: int
    nvidia_gpu: int = 0
    amd_gpu: int = 0
    intel_gpu: int = 0


@dataclass(frozen=True)
class IdleJobConfig:
    name: str
    count: int
    image: str
    resources: Resources
    command: list[str] = field(default_factory=list)
    args: list[str] = field(default_factory=list)
    image_pull_secret: str | None = None
    env: dict[str, str] = field(default_factory=dict)
    node_selector: dict[str, str] = field(default_factory=dict)


@dataclass(frozen=True)
class OrchestratorConfig:
    job_hostname_template: str
    job_fallback_hostname: str
    job_schedule_timeout_s: float = 300
    job_schedule_scale_up_timeout_s: float = 300
    is_http_ingress_secure: bool = True
    resource_pool_types: Sequence[ResourcePoolType] = ()
    resource_presets: Sequence[ResourcePreset] = ()
    allow_privileged_mode: bool = False
    allow_job_priority: bool = False
    pre_pull_images: Sequence[str] = ()
    idle_jobs: Sequence[IdleJobConfig] = ()

    @property
    def allow_scheduler_enabled_job(self) -> bool:
        for preset in self.resource_presets:
            if preset.scheduler_enabled:
                return True
        return False

    @property
    def tpu_resources(self) -> Sequence[TPUResource]:
        return tuple(
            resource.tpu for resource in self.resource_pool_types if resource.tpu
        )

    @property
    def tpu_ipv4_cidr_block(self) -> str | None:
        tpus = self.tpu_resources
        if not tpus:
            return None
        return tpus[0].ipv4_cidr_block


@dataclass
class PatchOrchestratorConfigRequest:
    job_hostname_template: str | None = None
    job_fallback_hostname: str | None = None
    job_schedule_timeout_s: float | None = None
    job_schedule_scale_up_timeout_s: float | None = None
    is_http_ingress_secure: bool | None = None
    resource_pool_types: Sequence[ResourcePoolType] | None = None
    resource_presets: Sequence[ResourcePreset] | None = None
    allow_privileged_mode: bool | None = None
    allow_job_priority: bool | None = None
    pre_pull_images: Sequence[str] | None = None
    idle_jobs: Sequence[IdleJobConfig] | None = None


@dataclass
class ARecord:
    name: str
    ips: Sequence[str] = ()
    dns_name: str | None = None
    zone_id: str | None = None
    evaluate_target_health: bool = False


@dataclass(frozen=True)
class DNSConfig:
    name: str = "not-used"
    a_records: Sequence[ARecord] = ()


@dataclass(frozen=True)
class EnergySchedulePeriod:
    # ISO 8601 weekday number (1-7)
    weekday: int
    start_time: time
    end_time: time

    @classmethod
    def create_full_day(cls, *, weekday: int, timezone: tzinfo) -> EnergySchedulePeriod:
        return cls(
            weekday=weekday,
            start_time=time.min.replace(tzinfo=timezone),
            end_time=time.max.replace(tzinfo=timezone),
        )


DEFAULT_ENERGY_SCHEDULE_NAME = "default"


@dataclass(frozen=True)
class EnergySchedule:
    name: str
    periods: Sequence[EnergySchedulePeriod] = ()
    price_per_kwh: Decimal = Decimal("0")

    @classmethod
    def create_default(
        cls, *, timezone: tzinfo, name: str = DEFAULT_ENERGY_SCHEDULE_NAME
    ) -> EnergySchedule:
        return cls(
            name=name,
            periods=[
                EnergySchedulePeriod.create_full_day(weekday=weekday, timezone=timezone)
                for weekday in range(1, 8)
            ],
        )

    def check_time(self, current_time: datetime) -> bool:
        return any(self._is_time_within_period(current_time, p) for p in self.periods)

    def _is_time_within_period(
        self, time_: datetime, period: EnergySchedulePeriod
    ) -> bool:
        time_ = time_.astimezone(period.start_time.tzinfo)
        return (
            period.weekday == time_.isoweekday()
            and period.start_time <= time_.timetz() < period.end_time
        )


@dataclass(frozen=True)
class EnergyConfig:
    co2_grams_eq_per_kwh: float = 0
    schedules: Sequence[EnergySchedule] = ()

    def get_schedule(self, name: str) -> EnergySchedule:
        return (
            self._get_schedule(name)
            or self._get_schedule(DEFAULT_ENERGY_SCHEDULE_NAME)
            or EnergySchedule.create_default(timezone=UTC)
        )

    def _get_schedule(self, name: str) -> EnergySchedule | None:
        for schedule in self.schedules:
            if schedule.name == name:
                return schedule
        return None

    @property
    def schedule_names(self) -> list[str]:
        return [schedule.name for schedule in self.schedules]


@dataclass(frozen=True)
class AppsConfig:
    apps_hostname_templates: list[str]
    app_proxy_url: URL


@dataclass(frozen=True)
class Cluster:
    name: str
    orchestrator: OrchestratorConfig
    storage: StorageConfig
    registry: RegistryConfig
    monitoring: MonitoringConfig
    secrets: SecretsConfig
    grafana: GrafanaConfig
    prometheus: PrometheusConfig
    disks: DisksConfig
    buckets: BucketsConfig
    apps: AppsConfig
    dns: DNSConfig = DNSConfig()
    ingress: IngressConfig = IngressConfig()
    energy: EnergyConfig = EnergyConfig()
    location: str | None = None
    logo_url: URL | None = None
    timezone: tzinfo = UTC
    created_at: datetime = field(
        default=datetime.min.replace(tzinfo=UTC), compare=False
    )


@dataclass(frozen=True)
class PatchClusterRequest:
    location: str | None = None
    logo_url: URL | None = None
    storage: StorageConfig | None = None
    registry: RegistryConfig | None = None
    orchestrator: PatchOrchestratorConfigRequest | None = None
    monitoring: MonitoringConfig | None = None
    secrets: SecretsConfig | None = None
    grafana: GrafanaConfig | None = None
    prometheus: PrometheusConfig | None = None
    disks: DisksConfig | None = None
    buckets: BucketsConfig | None = None
    timezone: ZoneInfo | None = None
    energy: EnergyConfig | None = None
    apps: AppsConfig | None = None
